from rxxxt.elements import HTMLFragment, Element, HTMLElement, HTMLVoidElement, UnescapedHTMLElement, CustomAttribute, VEl, El # noqa: F401
from rxxxt.component import Component, event_handler, HandleNavigate # noqa: F401
from rxxxt.page import default_page, PageBuilder # noqa: F401
from rxxxt.execution import Context # noqa: F401
from rxxxt.app import App # noqa: F401
from rxxxt.router import Router # noqa: F401
from rxxxt.state import local_state as local_state, gloabl_state as gloabl_state



from abc import ABC, abstractmethod
import base64
from datetime import datetime, timedelta, timezone
import hashlib
import inspect
from io import BytesIO
import json
import os
import secrets
from typing import Awaitable, Callable, Generic, Literal, TypeVar, cast, get_origin
from pydantic import TypeAdapter, ValidationError
import hmac

from rxxxt.component import Component

T = TypeVar("T")

class StateDescriptor(Generic[T]):
  _native_types = { bool, bytearray, bytes, complex, dict, float, frozenset, int, list, object, set, str, tuple }

  def __init__(self, is_global: bool, default_factory: Callable[[], T], state_name: str | None = None) -> None:
    self._is_global = is_global
    self._state_name = state_name
    self._default_factory = default_factory

    if default_factory in StateDescriptor._native_types or get_origin(default_factory) in StateDescriptor._native_types:
      self._val_type_adapter = TypeAdapter(default_factory)
    else:
      sig = inspect.signature(default_factory)
      self._val_type_adapter = TypeAdapter(sig.return_annotation)

  def __set_name__(self, owner, name):
    if self._state_name is None:
      self._state_name = name

  def __set__(self, obj, value):
    if not isinstance(obj, Component):
      raise TypeError("StateDescriptor used on non-component!")
    svalue = self._val_type_adapter.dump_json(value).decode("utf-8")
    obj.context.set_state(self._get_state_name(obj), svalue)

  def __get__(self, obj, objtype=None):
    if not isinstance(obj, Component):
      raise TypeError("StateDescriptor used on non-component!")

    svalue = obj.context.get_state(self._get_state_name(obj))
    if svalue is None: return self._default_factory()
    else: return cast(T, self._val_type_adapter.validate_json(svalue))

  def _get_state_name(self, comp: Component):
    if self._state_name is None: raise ValueError("state name is not set!")
    if self._is_global: return f"global;{self._state_name}"
    else: return f"#local;{comp.context.sid};{self._state_name}"

def local_state(default_factory: Callable[[], T], name: str | None = None):
  return StateDescriptor(False, default_factory, state_name=name)

def gloabl_state(default_factory: Callable[[], T], name: str | None = None):
  return StateDescriptor(True, default_factory, state_name=name)

CompressedState = dict[str, str | dict[str, str]]
CompressedStateAdapter = TypeAdapter(CompressedState)

def compress_state(state: dict[str, str]) -> CompressedState:
  prefixed_state: dict[str, dict[str, str]] = {}
  for k, v in state.items():
    sep_idx = k.rfind("!") + 1
    k1 = k[:sep_idx]
    k2 = k[sep_idx:]
    if k1 not in prefixed_state: prefixed_state[k1] = {}
    prefixed_state[k1][k2] = v

  compressed_state: CompressedState = {}
  for prefix, substate in prefixed_state.items():
    if len(substate) == 1:
      for suffix, value in substate.items():
        compressed_state[prefix + suffix] = value
    else:
      compressed_state[prefix] = substate
  return compressed_state

def decompress_state(compressed_state: CompressedState):
  state: dict[str, str] = {}
  for prefix, substate in compressed_state.items():
    if isinstance(substate, str): state[prefix] = substate
    else:
      for suffix, value in substate.items():
        state[prefix + suffix] = value
  return state

class StateResolverError(BaseException): pass

class StateResolver(ABC):
  @abstractmethod
  def create_token(self, data: dict[str, str], old_token: str | None) -> str | Awaitable[str]: pass
  @abstractmethod
  def resolve(self, token: str) -> dict[str, str] | Awaitable[dict[str, str]]: pass

class JWTStateResolver(StateResolver):
  def __init__(self, secret: bytes, max_age: timedelta | None = None, algorithm: Literal["HS256"] | Literal["HS384"] | Literal["HS512"] = "HS512") -> None:
    super().__init__()
    self.secret = secret
    self.algorithm = algorithm
    self.digest = { "HS256": hashlib.sha256, "HS384": hashlib.sha384, "HS512": hashlib.sha512 }[algorithm]
    self.max_age: timedelta = timedelta(days=1) if max_age is None else max_age

  def create_token(self, data: dict[str, str], old_token: str | None) -> str:
    payload = { "exp": int((datetime.now(tz=timezone.utc) + self.max_age).timestamp()), "data": compress_state(data) }
    stream = BytesIO()
    stream.write(JWTStateResolver.b64url_encode(json.dumps({
      "typ": "JWT",
      "alg": self.algorithm
    }).encode("utf-8")))
    stream.write(b".")
    stream.write(JWTStateResolver.b64url_encode(json.dumps(payload).encode("utf-8")))

    signature = hmac.digest(self.secret, stream.getvalue(), self.digest)
    stream.write(b".")
    stream.write(JWTStateResolver.b64url_encode(signature))
    return stream.getvalue().decode("utf-8")

  def resolve(self, token: str) -> dict[str, str] | Awaitable[dict[str, str]]:
    rtoken = token.encode("utf-8")
    sig_start = rtoken.rfind(b".")
    if sig_start == -1: raise StateResolverError("Invalid token format")
    parts = rtoken.split(b".")
    if len(parts) != 3: raise StateResolverError("Invalid token format")

    try: header = json.loads(JWTStateResolver.b64url_decode(parts[0]))
    except: raise StateResolverError("Invalid token header")

    if not isinstance(header, dict) or header.get("typ", None) != "JWT" or header.get("alg", None) != self.algorithm:
      raise StateResolverError("Invalid header contents")

    signature = JWTStateResolver.b64url_decode(rtoken[(sig_start + 1):])
    actual_signature = hmac.digest(self.secret, rtoken[:sig_start], self.digest)
    if not hmac.compare_digest(signature, actual_signature):
      raise StateResolverError("Invalid JWT signature!")

    payload = json.loads(JWTStateResolver.b64url_decode(parts[1]))
    if not isinstance(payload, dict) or not isinstance(payload.get("exp", None), int) or not isinstance(payload.get("data", None), dict):
      raise StateResolverError("Invalid JWT payload!")

    expires_dt = datetime.fromtimestamp(payload["exp"], timezone.utc)
    if expires_dt < datetime.now(tz=timezone.utc):
      raise StateResolverError("JWT expired!")

    try: compressed_state = CompressedStateAdapter.validate_python(payload["data"])
    except ValidationError as e: raise StateResolverError(e)
    return decompress_state(compressed_state)

  @staticmethod
  def b64url_encode(value: bytes | bytearray): return base64.urlsafe_b64encode(value).rstrip(b"=")
  @staticmethod
  def b64url_decode(value: bytes | bytearray): return base64.urlsafe_b64decode(value + b"=" * (4 - len(value) % 4))

def default_state_resolver():
  jwt_secret = os.getenv("JWT_SECRET", None)
  if jwt_secret is None: jwt_secret = secrets.token_bytes(64)
  else: jwt_secret = jwt_secret.encode("utf-8")
  return JWTStateResolver(jwt_secret)



from abc import abstractmethod
import asyncio
import base64
import inspect
import json
from typing import Annotated, Any, Callable, Coroutine, Generic, ParamSpec, TypeVar, get_args, get_origin
from pydantic import BaseModel, validate_call
from typing_extensions import Awaitable
from rxxxt.elements import CustomAttribute, Element, meta_element
from rxxxt.execution import Context, InputEvent
from rxxxt.helpers import to_awaitable
from rxxxt.node import Node

EHP = ParamSpec('EHP')
EHR = TypeVar('EHR')

class EventHandlerOptions(BaseModel):
  debounce: int | None = None
  throttle: int | None = None
  prevent_default: bool = False

class ClassEventHandler(Generic[EHP, EHR]):
  def __init__(self, fn:  Callable[EHP, EHR], options: EventHandlerOptions) -> None:
    self.fn = fn
    self.options = options
  def __get__(self, instance, owner): return InstanceEventHandler(self.fn, self.options, instance)
  def __call__(self, *args: EHP.args, **kwargs: EHP.kwargs) -> EHR: raise RuntimeError("The event handler can only be called when attached to an instance!")

class InstanceEventHandler(ClassEventHandler, Generic[EHP, EHR], CustomAttribute):
  def __init__(self, fn: Callable[EHP, EHR], options: EventHandlerOptions, instance: Any) -> None:
    super().__init__(validate_call(fn), options)
    if not isinstance(instance, Component): raise ValueError("The provided instance must be a component!")
    self.instance = instance

  def __call__(self, *args: EHP.args, **kwargs: EHP.kwargs) -> EHR: return self.fn(self.instance, *args, **kwargs)

  def get_key_value(self, original_key: str):
    if not original_key.startswith("on"): raise ValueError("Event handler must be applied to an attribute starting with 'on'.")
    if self.instance.context is None: raise ValueError("The instance must have a context_id to create an event value.")
    v = base64.b64encode(json.dumps({
      "context_id": self.instance.context.sid,
      "handler_name": self.fn.__name__,
      "param_map": self._get_param_map(),
      "options": self.options.model_dump(exclude_defaults=True)
    }).encode("utf-8")).decode("utf-8")
    return (f"rxxxt-on-{original_key[2:]}", v)

  def _get_param_map(self):
    param_map: dict[str, str] = {}
    sig = inspect.signature(self.fn)

    for i, (name, param) in enumerate(sig.parameters.items()):
      if i == 0: continue # skip self

      if get_origin(param.annotation) is Annotated:
        args = get_args(param.annotation)
        metadata = args[1:]

        if len(metadata) < 1:
          raise ValueError(f"Parameter '{name}' is missing the second annotation.")

        param_map[name] = metadata[0]
      else:
        raise TypeError(f"Parameter '{name}' must be of type Annotated.")

    return param_map

def event_handler(**kwargs):
  options = EventHandlerOptions.model_validate(kwargs)
  def _inner(fn) -> ClassEventHandler: return ClassEventHandler(fn, options)
  return _inner

class HandleNavigate(CustomAttribute):
  def __init__(self, location: str) -> None:
    super().__init__()
    self.location = location

  def get_key_value(self, original_key: str) -> tuple[str, str]:
    return (original_key, f"window.rxxxt.navigate('{self.location}');")

class Component(Element):
  def __init__(self) -> None:
    super().__init__()
    self.context: Context
    self._bg_tasks: list[asyncio.Task] = []

  @abstractmethod
  def render(self) -> Element | Awaitable[Element]: ...

  def add_background_task(self, a: Coroutine):
    if not self.context.config.persistent:
      a.close()
      return
    self._bg_tasks.append(asyncio.create_task(a))
  def request_update(self): self.context.request_update()

  async def lc_init(self, context: Context) -> None:
    self.context = context
    await self.on_init()
  async def lc_destroy(self) -> None:
    await self.on_before_destroy()
    if len(self._bg_tasks) > 0:
      for t in self._bg_tasks: t.cancel()
      try: await asyncio.wait(self._bg_tasks)
      except asyncio.CancelledError: pass
      self._bg_tasks.clear()
    await self.on_after_destroy()
  async def lc_handle_event(self, event: dict[str, int | float | str | bool]):
    handler_name = event.pop("$handler_name", None)
    if isinstance(handler_name, str):
      handler = getattr(self, handler_name, None) # NOTE: this is risky!!
      if isinstance(handler, InstanceEventHandler):
        await to_awaitable(handler, **event)

  async def on_init(self) -> None: ...
  async def on_before_destroy(self) -> None: ...
  async def on_after_destroy(self) -> None: ...

  def tonode(self, context: Context) -> 'Node': return ComponentNode(context, self)

class ComponentNode(Node):
  def __init__(self, context: Context, element: Component) -> None:
    super().__init__(context, [])
    self.element = element

  async def expand(self):
    if len(self.children) > 0:
      raise ValueError("Can not expand already expanded element!")

    await self.element.lc_init(self.context)
    await self._render_inner()

  async def update(self):
    for c in self.children: await c.destroy()
    self.children.clear()
    await self._render_inner()

  async def handle_events(self, events: list[InputEvent]):
    for e in events:
      if self.context.sid == e.context_id:
        await self.element.lc_handle_event(dict(e.data))
    await super().handle_events(events)

  async def destroy(self):
    for c in self.children: await c.destroy()
    self.children.clear()

    await self.element.lc_destroy()
    self.context.unregister()

  async def _render_inner(self):
    inner = await to_awaitable(self.element.render)
    self.children.append(meta_element(self.context.sid, inner).tonode(self.context.sub("inner")))
    await self.children[0].expand()



import asyncio
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
import hashlib
import re
from typing import Literal

from pydantic import BaseModel, TypeAdapter, field_serializer, field_validator

ContextStackKey = str | int
ContextStack = tuple[ContextStackKey, ...]

class SetCookieOutputEvent(BaseModel):
  event: Literal["set-cookie"] = "set-cookie"
  name: str
  value: str | None = None
  expires: datetime | None = None
  path: str | None = None
  max_age: int | None = None
  secure: bool | None = None
  http_only: bool | None = None
  domain: str | None = None

  @field_validator('name')
  @classmethod
  def validate_name(cls, value: str):
    if not re.match(r'^[^=;, \t\n\r\f\v]+$', value): raise ValueError("Invalid cookie name")
    return value

  @field_validator('value', "domain")
  @classmethod
  def validate_value(cls, value: str | None):
    if value is None: return None
    if not re.match(r'^[^;, \t\n\r\f\v]+$', value): raise ValueError("Invalid value.")
    return value

  @field_validator('path')
  @classmethod
  def validate_path(cls, value: str | None):
    if value is None: return None
    if not re.match(r'^[^\x00-\x20;,\s]+$', value): raise ValueError("Invalid path value")
    return value

  @field_serializer('expires', when_used='json')
  def seriliaze_expires(self, value: datetime | None): return None if value is None else value.isoformat()

  def to_set_cookie_header(self):
    parts: list[str] = [f"{self.name}={self.value}"]
    if self.path is not None: parts.append(f"path={self.path}")
    if self.expires is not None: parts.append(f"expires={self.expires.strftime("%a, %d %b %G %T %Z")}")
    if self.max_age is not None: parts.append(f"max-age={self.max_age}")
    if self.domain is not None: parts.append(f"domain={self.domain}")
    if self.secure: parts.append("secure")
    if self.http_only: parts.append("httponly")
    return ";".join(parts)

class UseWebsocketOutputEvent(BaseModel):
  event: Literal["use-websocket"] = "use-websocket"
  websocket: bool

class NavigateOutputEvent(BaseModel):
  location: str
  event: Literal["navigate"] = "navigate"

class ContextInputEvent(BaseModel):
  context_id: str
  data: dict[str, int | float | str | bool]

InputEvent = ContextInputEvent
OutputEvent = SetCookieOutputEvent | NavigateOutputEvent | UseWebsocketOutputEvent
HeaderTypeAdapter = TypeAdapter(dict[str, list[str]])

class State:
  def __init__(self, update_event: asyncio.Event) -> None:
    self._state: dict[str, str] = {}
    self._state_subscribers: dict[str, set[ContextStack]] = {}
    self._pending_updates: set[ContextStack] = set()
    self._output_events: list[OutputEvent] = []
    self._update_event = update_event

  @property
  def data(self): return dict(self._state)

  def get_state(self, context_id: ContextStack, key: str) -> str | None:
    self._state_subscribers.setdefault(key, set()).add(context_id)
    return self._state.get(key)

  def set_state(self, key: str, value: str | None):
    if value != self._state.get(key, None):
      for cid in self._state_subscribers.get(key, []): self.request_update(cid)
      if value is None: self._state.pop(key, None)
      else: self._state[key] = value

  def add_output_event(self, event: OutputEvent): self._output_events.append(event)
  def pop_output_events(self):
    res = self._output_events
    self._output_events = []
    return res
  def pop_updates(self):
    res = self._pending_updates
    self._pending_updates = set()
    self._set_update_event()
    return res
  def request_update(self, cid: ContextStack):
    self._pending_updates.add(cid)
    self._set_update_event()

  def unregister(self, cid: ContextStack): # cleanup for lower memory usage
    if cid in self._pending_updates: self._pending_updates.remove(cid)
    self._set_update_event()

    for subs in self._state_subscribers.values():
      if cid in subs: subs.remove(cid)

  def _set_update_event(self):
    if len(self._pending_updates) > 0: self._update_event.set()
    else: self._update_event.clear()

@dataclass(frozen=True)
class ContextConfig:
  persistent: bool

class Context:
  def __init__(self, state: State, config: ContextConfig, stack: ContextStack) -> None:
    self._stack: ContextStack = stack
    self._state = state
    self._config = config

  @property
  def config(self): return self._config

  @property
  def id(self): return self._stack

  @cached_property
  def sid(self):
    hasher = hashlib.sha256()
    for k in self._stack:
      if isinstance(k, str): k = k.replace(";", ";;")
      else: k = str(k)
      hasher.update((k + ";").encode("utf-8"))
    return hasher.hexdigest()

  @property
  def location(self):
    res = self._state.get_state(self._stack, "location")
    if res is None: raise ValueError("No location!")
    else: return res

  @property
  def path(self): return self.location.split("?")[0]

  @property
  def query_string(self):
    parts = self.location.split("?")
    if len(parts) < 2: return None
    else: return parts[1]

  @property
  def headers(self): return HeaderTypeAdapter.validate_json(self.get_state("headers") or "")

  @property
  def cookies(self) -> dict[str, str]:
    values = self.headers.get("cookie", [])
    if len(values) == 0: return {}
    result: dict[str, str] = {}
    for cookie in values[0].split(";"):
      try:
        eq_idx = cookie.index("=")
        result[cookie[:eq_idx]] = cookie[(eq_idx + 1):]
      except ValueError: pass
    return result

  def sub(self, key: ContextStackKey): return Context(self._state, self._config, self._stack + (key,))
  def replace_index(self, key: str):
    if isinstance(self._stack[-1], int): return Context(self._state, self._config, self._stack[:-1] + (key,))
    raise ValueError("No index to replace!")

  def set_state(self, key: str, value: str): self._state.set_state(key, value)
  def get_state(self, key: str): return self._state.get_state(self.id, key)

  def request_update(self): self._state.request_update(self._stack)
  def unregister(self): self._state.unregister(self._stack)

  def navigate(self, location: str):
    self.set_state("location", location)
    self._state.add_output_event(NavigateOutputEvent(location=location))
  def use_websocket(self, websocket: bool = True): self._state.add_output_event(UseWebsocketOutputEvent(websocket=websocket))
  def set_cookie(self, name: str, value: str, expires: datetime | None = None, path: str | None = None,
                secure: bool | None = None, http_only: bool | None = None, domain: str | None = None, max_age: int | None = None):
    self._state.add_output_event(SetCookieOutputEvent(name=name, value=value, expires=expires, path=path, secure=secure, http_only=http_only, domain=domain, max_age=max_age))
  def delete_cookie(self, name: str):
    self._state.add_output_event(SetCookieOutputEvent(name=name, max_age=-1))



import asyncio
from io import StringIO
import unittest

from rxxxt.elements import Element, El
from rxxxt.execution import Context, ContextConfig, State

async def render_element(el: Element):
  state = State(asyncio.Event())
  context = Context(state, ContextConfig(persistent=False), ("root",))
  node = el.tonode(context)
  await node.expand()
  io = StringIO()
  node.write(io)
  await node.destroy()
  return io.getvalue()

class TestRender(unittest.IsolatedAsyncioTestCase):
  async def test_div(self):
    text = await render_element(El.div(content=["Hello World!"]))
    self.assertEqual(text, "<div>Hello World!</div>")


if __name__ == "__main__":
  unittest.main()



from abc import ABC, abstractmethod
import html
import logging
from typing import Protocol
from rxxxt.execution import Context
from rxxxt.node import ElementNode, FragmentNode, Node, TextNode, VoidElementNode

class CustomAttribute(ABC):
  @abstractmethod
  def get_key_value(self, original_key: str) -> tuple[str, str | None]: ...

class Element(ABC):
  @abstractmethod
  def tonode(self, context: Context) -> Node: ...

ElementContent = list[Element | str]
HTMLAttributeValue = str | bool | int | float | CustomAttribute | None

def element_content_to_nodes(context: Context, content: ElementContent):
  nodes: list[Node] = []
  for idx, c in enumerate(content):
    scontext = context.sub(idx)
    if isinstance(c, Element): nodes.append(c.tonode(scontext))
    elif isinstance(c, str): nodes.append(TextNode(scontext, html.escape(c)))
    else: raise ValueError("Invalid child!")
  return nodes

class HTMLFragment(Element):
  def __init__(self, content: ElementContent) -> None:
    super().__init__()
    self._content = content

  def tonode(self, context: Context) -> Node:
    return FragmentNode(context, element_content_to_nodes(context, self._content))

class HTMLVoidElement(Element):
  def __init__(self, tag: str, attributes: dict[str, HTMLAttributeValue]) -> None:
    super().__init__()
    self._tag = tag
    self._attributes: dict[str, str | None] = {}
    for k, v in attributes.items():
      if isinstance(v, CustomAttribute): k, v = v.get_key_value(k)
      elif isinstance(v, bool):
        if not v: continue
        v = None
      elif isinstance(v, (int, float)): v = str(v)
      self._attributes[k] = v

  def tonode(self, context: Context) -> 'Node':
    return VoidElementNode(context, self._tag, self._attributes)

class HTMLElement(HTMLVoidElement):
  def __init__(self, tag: str, attributes: dict[str, HTMLAttributeValue], content: ElementContent) -> None:
    super().__init__(tag, attributes)
    self._content = content

  def tonode(self, context: Context) -> 'Node':
    return ElementNode(context, self._tag, self._attributes, element_content_to_nodes(context, self._content))

class KeyedElement(Element):
  def __init__(self, key: str, element: Element) -> None:
    super().__init__()
    self._key = key
    self._element = element

  def tonode(self, context: Context) -> 'Node':
    try: context = context.replace_index(self._key)
    except ValueError as e: logging.debug(f"Failed to replace index with key {self._key}", e)
    return self._element.tonode(context)

class UnescapedHTMLElement(Element):
  def __init__(self, text: str) -> None:
    super().__init__()
    self._text = text

  def tonode(self, context: Context) -> 'Node': return TextNode(context, self._text)

class CreateHTMLElement(Protocol):
  def __call__(self, content: list[Element | str] = [], key: str | None = None, **kwargs: HTMLAttributeValue) -> Element: ...

class _El(type):
  def __getitem__(cls, name: str) -> CreateHTMLElement:
    def _inner(content: ElementContent = [], key: str | None = None, **kwargs: HTMLAttributeValue):
      el = HTMLElement(name, attributes={ k.lstrip("_"): v for k,v in kwargs.items() }, content=content)
      if key is not None: el = KeyedElement(key, el)
      return el
    return _inner
  def __getattribute__(cls, name: str): return cls[name]

class El(metaclass=_El): ...

class CreateHTMLVoidElement(Protocol):
  def __call__(self, **kwargs: HTMLAttributeValue) -> HTMLVoidElement: ...

class _VEl(type):
  def __getitem__(cls, name: str) -> CreateHTMLVoidElement:
    def _inner(**kwargs: HTMLAttributeValue) -> HTMLVoidElement:
      return HTMLVoidElement(name, attributes={ k.lstrip("_"): v for k,v in kwargs.items() })
    return _inner
  def __getattribute__(cls, name: str): return cls[name]

class VEl(metaclass=_VEl): ...

class ElementFactory(Protocol):
  def __call__(self) -> Element: ...

def meta_element(id: str, inner: Element): return HTMLElement("rxxxt-meta", {"id":id}, [inner])
